#include "gpio.h"

static const struct gpio::port ports[] =
{
#ifdef STM32F030x6
	{GPIOA,&(RCC->AHBENR),RCC_AHBENR_GPIOAEN},
	{GPIOB,&(RCC->AHBENR),RCC_AHBENR_GPIOBEN},
	{GPIOF,&(RCC->AHBENR),RCC_AHBENR_GPIOFEN}
#else
#error "MCU not supported"
#endif
};

void gpio::mode(pins pin, pinMode selMode, pullDirection selPull, speed selSpeed)
{
	uint8_t pinNumber = pin & 0xf;
	GPIO_TypeDef *port = (GPIO_TypeDef *)(pin & 0xfffffff0);
	// Enable GPIO Port
	for(uint8_t i = 0; i < (sizeof(ports)/sizeof(ports[0])); i ++) {
		if (ports[i].GPIO == port) {
			*(ports[i].reg) |= ports[i].bit;
			break;
		}
	}
	port->MODER = (port->MODER
		       & ~(0b11<<(pinNumber * 2)))
		| ((selMode&0b11)<<(pinNumber * 2));
	port->PUPDR = (port->PUPDR
		       & ~(0b11<<(pinNumber * 2)))
		| (selPull<<(pinNumber * 2));
	port->OSPEEDR = (port->OSPEEDR
			 & ~(0b11<<(pinNumber * 2)))
		| (selSpeed<<(pinNumber * 2));
	port->OTYPER = (port->OTYPER
			& ~(0b1<<pinNumber))
		| (((selMode & 0b100)>>2)<<pinNumber);
}

void gpio::high(pins pin)
{
	uint8_t pinNumber = pin & 0xf;
	GPIO_TypeDef *port = (GPIO_TypeDef *)(pin & 0xfffffff0);
	port->BSRR = (1<<pinNumber);
}

void gpio::low(pins pin)
{
	uint8_t pinNumber = pin & 0xf;
	GPIO_TypeDef *port = (GPIO_TypeDef *)(pin & 0xfffffff0);
	port->BRR = (1<<pinNumber);
}

void gpio::set(pins pin, bool state)
{
	uint8_t pinNumber = pin & 0xf;
	GPIO_TypeDef *port = (GPIO_TypeDef *)(pin & 0xfffffff0);
	port->BSRR = (state<<pinNumber) | (1<<(pinNumber+16));
}

void gpio::toggle(pins pin)
{
	uint8_t pinNumber = pin & 0xf;
	GPIO_TypeDef *port = (GPIO_TypeDef *)(pin & 0xfffffff0);
	port->ODR ^= (1<<pinNumber);
}

uint8_t gpio::read(pins pin)
{
	uint8_t pinNumber = pin & 0xf;
	GPIO_TypeDef *port = (GPIO_TypeDef *)(pin & 0xfffffff0);
	return ((port->IDR & (1<<pinNumber)) == (uint32_t)(1<<pinNumber));
}

void gpio::af(pins pin, AF selected)
{
	uint8_t pinNumber = pin & 0xf;
	GPIO_TypeDef *port = (GPIO_TypeDef *)(pin & 0xfffffff0);
	// The 3rd bit is cleared, because it's only used in deciding
	// which register to use
	uint8_t afind = (pinNumber & 0b111) * 4;
	// Set the af bits for the pin
	// The right shift finds if the highest bit in the pin nibble
	// is high, which decides what register should be used
	port->AFR[pinNumber >> 4] =
		//Clear the 4 bits defining the  af of the gpio pin
		(port->AFR[pinNumber >> 3] & ~(0b1111 << (afind)))
		// Set the neccessary bits
		| (selected << (afind));
}
