#ifndef __USART_H
#define __USART_H
#include "stm32f0xx.h"
#include <stdint.h>
namespace USART {
#ifdef USART_USE_FIFO
	//Ringbuffer length
	const uint32_t USART_FifoBufferLength = 512;
#endif
	typedef enum  {
		USART_ReturnSuccess = 0,
		USART_ReturnFailed = 1,
	}USART_Status;
	class USART {
	public:
		USART(void);
		~USART(void);
		void Init(uint32_t baudrate = 9600, USART_TypeDef *UART_InstanceParameter = USART1);
		USART_Status TransmitByte(uint8_t data);
		uint8_t ReceiveByte(void);
		/* Checks if a byte is in the stm32 usart receive fifo */
		bool ReceiveAvailable(void);
		USART_Status PrintString(const char myString[]);
		uint8_t ReadString(char myString[], uint8_t maxLength);
		void PrintByte(uint8_t byte);
		USART_Status PrintWord(uint16_t word);
		void PrintBinaryByte(uint8_t byte);
		void InterruptHandler(void);
		/* Call this from the interrupt routine of the USART */
		uint8_t ReadHexByte(void);
		void PrintHexByte(uint8_t byte);
		uint8_t GetNumber(void);
		/* Returns the free Size of the Fifo */
		uint32_t FifoSize(void);
	private:
		USART_TypeDef *Instance;
#ifdef USART_USE_FIFO
		//Fifo struct
		struct USART_FifoBufferTypeDef {
			char buffer[USART_FifoBufferLength];
			//Data gets written here
			uint32_t head;
			//Data gets read here
			uint32_t tail;
		};
		//Lets make a global Fifo Struct
		volatile USART_FifoBufferTypeDef FifoBuffer;

		/* Inits the Fifo */
		void FifoInit(void);
		/* Puts a value in the FIFO, fails when there isn't enough space */
		USART_Status FifoPut(uint8_t data);
		/* Gets a value from the FIFO, fails when there isn't another value*/
		USART_Status FifoGet(uint8_t &data);
#endif
		char NibbleToHexCharacter(uint8_t nibble);
		uint8_t HexCharacterToNibble(char hex);
	};

	/* Inits USART1 with as a standard USART with 115200 Baud and the Transmit complete Interrupt
	 * (Still needs to be enable in the Core)
	 * Warning: Doesn't initialize the GPIO, doesn't activate the clock and doesn't activate the interrupt in the core (which is needed)
	 */

	/* Conigures PA9 and PA10 in USART AF Mode and activates USART1 Clock
	 * Warning: If you don't use the STM32f030 series, USART1 or want to use another pair of pins,
	 * you have to use your own init for the GPIO Pins
	 */
	void SpecificInitDefault(void);


}
#endif
